//
// Classe per rappresentare una Pila.
// Utilizzo di templates ed ereditarieta`.
//

#ifndef STACK_HDR
#define STACK_HDR

// Classi base

class BaseNode {
  friend class BaseStack;	// BaseStack puo` accedere alla parte privata

  BaseNode *next;		// Puntatore al prossimo elemento
};


class BaseStack {
  BaseNode *head;		// Puntatore ai dati
  int len;			// La lunghezza della Pila

public:
  BaseStack();			// Costruttore
  ~BaseStack();			// Distruttore

  BaseNode *Read();		// Restituisce l'elemento in testa  alla Pila
  void Push(BaseNode *d);	// Aggiunge un elemento in testa alla Pila
  BaseNode *Pop();		// Rimuove l'elemento in testa alla Pila (e lo
				// restituisce)
  int IsEmpty()			// La pila e` vuota?
    {return(len == 0); }
};



// Classi Derivate

// La classe Node e` una estensione della classe BaseNode, aggiungendo alla
// sua parte privata una variabile di tipo MyNType.
template<class MyNType>class Node: private BaseNode {
  friend class Stack<MyNType>;	// Stack<MyNType> deve manipolare data

  MyNType data;

public:
  // Copy constructor
  Node(const MyNType &d):
    data(d)
    {}
};


// La classe Stack estende la classe BaseStack, ridefinendo le sue funzioni in
// modo gestire dati di qualunque tipo.
template<class MyType> class Stack: private BaseStack {
public:
  ~Stack();
  MyType Read();		// Restituisce l'elemento in testa  alla Pila
  void Push(const MyType &d);	// Aggiunge un elemento in testa alla Pila
  MyType Pop();			// Rimuove l'elemento in testa alla Pila
  int IsEmpty()			// La pila e` vuota?
    {return(BaseStack::IsEmpty()); }
};



// Restituisce l'elemento in testa alla Pila
template<class MyType> MyType Stack<MyType>::Read()
{
  // Usiamo la funzione della classe BaseStack per accedere all'elemento in
  // testa alla Pila.
  // Il puntatore che ci restituisce e` ad un oggetto di tipo NodeBase, ma noi
  // sappiamo che in realta` e` di tipo Node<MyType> e quindi possiamo fare
  // un cast sicuri di non avere problemi.
  Node<MyType> *npntr = (Node<MyType> *)BaseStack::Read();


  return(npntr->data);
}



// Aggiunge un elemento in testa alla Pila
template<class MyType> void Stack<MyType>::Push(const MyType &d)
{
  // Creo un oggetto di tipo Node<MyType>
  Node<MyType> *nptr = new Node<MyType>(d);

  // Lo inserisco nella Pila utilizzando la classe StackBase. Non servono cast.
  BaseStack::Push(nptr);
}


// Rimuove l'elemento in testa alla Pila (e restituisce il suo campo
// informativo).
template<class MyType> MyType Stack<MyType>::Pop()
{
  // Come per la Read possimo effettuare il cast senza problemi.
  Node<MyType> *npntr = (Node<MyType> *)BaseStack::Pop();

  // Salvo il dato appena recuperato
  MyType tmp = npntr->data;

  // Visto che e` la classe Stack che ha allocato un oggetto, sara` questa
  // classe a deallocarlo. Notare che non e` necessario che i distruttori
  // siano virtuali visto che accedo attraverso un puntatore all'oggetto del
  // tipo giusto.
  delete npntr;

  return(tmp);
}


// Distruttore per il tipo di dato Stack.
template<class MyType> Stack<MyType>::~Stack()
{
  while (!IsEmpty()) {
    Pop();
  }
}


#endif
